/*******************************************************************************/
/*                                                                             */
/*                                                                             */
/*                      rsgen_gf_tables.c                                      */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*  Decription:                                                                */
/*     Generates a^n, 1/a^n, (a^n)^basis, tables for   GF(2^n)                 */
/*                                                                             */
/*      These tables are used for multiplication, inverse of a^n               */
/*      The power table is used if the basis of the generator                  */
/*      polynomial (j0) is                                                     */
/*                                                                             */
/*      ROM images for FPGAs  are also generated if required                   */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */ 
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/

#include <stdio.h>
#include "rsccg_proto.h"
//
//
//#define   C_MEM_ARRAY     0
//#define   XLX_VERILOG_SIM 2
//#define   XLX_VHDL_SIM    3
#define VHDL     0 
#define VLOG     1 
#define C        2 
//
// Work space arrays
//
unsigned char poly[256] ;            // Workspace for factoring GF feild generator polynomial 
unsigned char poly_table[256] ;      // a^n table         
unsigned char inv_poly_table[256] ;  // 1/a^n table      
unsigned char pow_poly_table[256] ;  // (a^n)^basis table 
unsigned char log_poly_table[256] ;  // n = log(a^n)     
//
//
//
//   rsgen_tables
//
//   input: 
//      gf_m    - Galois feild represented as 2^n (2^3 = 8)
//
//      base_j0 - Generator polynomial base
//
//      gen_mif
//         1 = generate Altera Memory Initialization file
//         2 = generate Xilinx Memory Initialization file
//         3 = generate Xilinx VHDL Simulation Initialation file
//                        
//      gen_txtfile
//         output tables to readable textfile
//
void rsgen_gf_tables(unsigned char *gf_feild_poly, int gf_m, int base_j0, int dbg)
{
unsigned char i, inverse, l ;
int n, p, r, j, gf_feild ;

   // Compute 2^gf_m
   gf_feild = 1 ;
   for (n=0; n < gf_m; n++) gf_feild = 2*gf_feild ;
    
   // Compute a^n
   if (dbg) printf("Computing A^n Table\n") ;
   for (n=0 ; n < gf_feild; n++) {
        r = factor_poly(gf_feild_poly, gf_m, gf_feild, n ) ;
        poly_table[n] = r ;
   }

   //
   // Compute Inverses of a^n (1/a^n)
   //
   if (dbg) printf("\n Computing Inverse A^n Table\n") ;
   inv_poly_table[0] = 0x01 ;
   for (n=0 ; n< gf_feild ; n++) {
       i = poly_table[n] ;
       j = (gf_feild-1) - n ;
       inverse = poly_table[j] ;
       inv_poly_table[i] = inverse ;
       if (dbg) printf("n=%d i=%2x j=%d inv=%2x\n",n,i,j,inverse) ;
   }

   //
   // Compute (a^n)^base where base is greater than 1
   // This lookup table is used for solving the Key equation when b > 1 
   //
   if (dbg) printf("\n Computing Power (A^n)^B Table\n") ;
   if (base_j0 > 0) {
       for (n=0 ; n < gf_feild ; n++) {
           i = poly_table[n] ;
           p = (n*(base_j0-1))%(gf_feild-1) ; /* power is 255 (GF2^8) x Base */
           j = poly_table[p] ;
           pow_poly_table[n] = j ;
           if (dbg) printf("n=%d i=%2x p=%d j=%2x \n",n,i,p,j) ;
       }
   }
   
   
   //
   // Compute n = log(a^n)
   // Used to find n from a 
   //
   if (dbg) printf("\n Computing log(A^N) Table\n") ;
   for (n=0 ; n < gf_feild; n++) {
       l = poly_table[n] ;
       log_poly_table[l] = n ;
       if (dbg) printf("n=%d l=%2x \n",n,l) ;
   }
   
   
}
//
//
//  factor_poly()
//     inputs: 
//         p - pointer to polynomial array 
//         gf_deg - degree of P(x)
//         gf_2n - Galois feild 2^n 
//         n - degree of x^n to factor with polynomial P(x)
//
//     output:
//         a^n in GF(2^n)
//
unsigned char factor_poly(unsigned char *p, int gf_deg, int gf_2n, int n )
{
int i, j;
char  k ;

 for ( i=0; i<256; i++)  /* Zero Poly Array */
    poly[i] = '0' ;

 poly[n] = '1' ;   /* Set bit position of poly to 2^n */

 //
 // Division of 2^n by primitive polynomial prim_poly
 //
 i = 255 ;
 while ( poly[i] == '0') i--  ; /* Find first '1' NOTE: This is redundant code */
 j = i ;  /* Save begining of first element of poly */

 while ( j >= gf_deg) {

 /*printf ("poly array first 1 element = %2d \n", i ) ; */

 /* Subtract prim_poly from poly */
 for ( k=0 ; k <= gf_deg; k++ ) {
    if (poly[i] == '1' && p[gf_deg-k] == '1')
        poly[i] = '0' ;
    else if (poly[i] == '0' && p[gf_deg-k] == '1')
        poly[i] = '1' ;
    else if (poly[i] == '1' && p[gf_deg-k] == '0')
        poly[i] = '1' ;
    i-- ;
 }

 i = 255 ;
 while ( poly[i] == '0') i--  ; /* Find first '1' */
 j = i  ;  /* Save begining of first element of poly */

} 

/* printf ("Division Complete\n") ; */
/*
  for ( i=0 ; i <= 10 ; i++) {
    printf("%c ", poly[10 - i] ) ;
  } 
*/
//
// Since prim_poly is bit reversed the result is also.
// We need to switch the result bits back to MSB..LSB
//
  j = 0 ;
  for ( i=0 ; i < 8 ; i++) {
    if (poly[7-i] == '1')
      k = 1 ;
    else if (poly[7-i] == '0')
      k = 0 ;

    j =  (j << 1) | k ;
  } 
  return( j ) ;
}

