/*******************************************************************************/
/*                                                                             */
/*            gen_rs_mul.c                                                     */
/*                                                                             */
/*    Generate RS multipliers from RS generator                                */
/*    polynomial                                                               */
/*                                                                             */
/* Copyright (C) 2003 John S. Willy                                            */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/* input:                                                                      */
/*     Pointer to array of RS generator coefficients                           */    
/*                                                                             */
/*     VHDL/Verilog                                                            */
/*                                                                             */
/* output:                                                                     */
/*     Scalar multiplier for each term in generator polynomial                 */
/*     Scalar multipliers for chien search algorithm a^0,a^1                   */
/*     a^n.                                                                    */
/*                                                                             */
/* returns:                                                                    */
/*     VOID                                                                    */
/*                                                                             */
/* comments:                                                                   */
/*        Polynomial is in reverse order [0] = A*x^0                           */
/*                                       [1] = B*x^1                           */
/*                                       [n] =  *x^n                           */
/*        each element in array represents coefficient of x^n                  */
/*                                                                             */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */ 
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/
#include <stdio.h>

#define VHDL     0 
#define VLOG     1 
#define C        2 

extern unsigned char poly_table[] ;   // RS code table 
extern unsigned char log_poly_table[] ; 
extern unsigned int p_wsp ;           // RS Generator Polynomial

/* locals  */
unsigned char mul_matrix[8] ;
//
// Generate Reed-Solomon hardware scalar multiplier - multiplies
// a*reed solomon scalar constant in GF(2^n)
//
// gen_rs_scalar_mul()
//
// input:
//     coeff: RS Generator polynomial coeffiecent or a^n
//     gf_m : Feild size (2^m)
//     hdl_type : HDL type, 0 = vhdl, 1 = verilog, 2 = C
//
// output:
//     HDL code
//
// returns:
//     void
//
//     
void gen_rs_scalar_mul(FILE *stream, unsigned int coeff, unsigned int gf_m, unsigned int gf_field, unsigned int hdl_type)
{

int i, j ;
int a, c, bit_cnt ;

unsigned char bit_mask ;

   for (i=0; i< 8; i++) {
          mul_matrix[i] = 0 ;
   }



   a = (unsigned int)log_poly_table[coeff]%(gf_field-1) ;
   c = a ;

   switch (hdl_type) {
       case C:
            if(coeff != 0) {
               fprintf(stream, "//\n") ;
               fprintf(stream, "// c * a^%d\n", c) ;
               fprintf(stream, "//\n") ;
               fprintf(stream, "unsigned char gfmul_%d( a )\n", c) ;
               fprintf(stream, "unsigned char a ;\n") ;
               fprintf(stream, "{\n") ;
               fprintf(stream, "int i ;\n") ;
               fprintf(stream, "unsigned char c[8], gf_mul[8], r ;\n") ;
               fprintf(stream, "\n") ;
               fprintf(stream, "  for ( i=0; i< %d; i++) {  /* Convert to single bit array for multiply */\n" , gf_m) ;
               fprintf(stream, "    c[i] = a & 0x01 ;\n") ;
               fprintf(stream, "    a = a >> 1 ;\n") ;
               fprintf(stream, "  }\n\n") ; }
           else {
               fprintf(stream, "//\n") ;
               fprintf(stream, "// c * a^%d\n", c) ;
               fprintf(stream, "//\n") ;
               fprintf(stream, "unsigned char gf_mul( a ) %d\n", c) ;
               fprintf(stream, "unsigned char a ;\n") ;
               fprintf(stream, "{\n") ;
               fprintf(stream, "\n") ;
               fprintf(stream, "return (a) ;\n") ;
               fprintf(stream, "}\n\n") ;
               return ;
           }
           break ;
       case VLOG : 
           if ( coeff !=0 ) {
               fprintf(stream, "//\n") ;
               fprintf(stream, "// c * a^%d\n", c) ;
               fprintf(stream, "//\n") ;
               fprintf(stream, "function[%d:0] gfmul_%d ;\n", gf_m-1, c) ;
               fprintf(stream, "input[%d:0] c ;\n", gf_m-1);
               fprintf(stream, "begin\n") ; }
           else {
               fprintf(stream, "//\n") ;
               fprintf(stream, "// c * a^%d\n", c) ;
               fprintf(stream, "//\n") ;
               fprintf(stream, "function[%d:0] gfmul_%d ;\n", gf_m-1, c) ;
               fprintf(stream, "input[%d:0] c;\n", gf_m-1);
               fprintf(stream, "end\n") ;
               fprintf(stream, "endfunction\n\n") ;
               return ;
           }
           break;
       case VHDL :
           if ( coeff !=0 ) {
               fprintf(stream, "--\n") ;
               fprintf(stream, "-- c * a^%d\n", c) ;
               fprintf(stream, "--\n") ;
               fprintf(stream, "function gfmul_%d(c : std_logic_vector(%d downto 0)) return std_logic_vector is\n", c, gf_m-1) ;
               fprintf(stream, "variable gfmul_%d : std_logic_vector(%d downto 0) ;\n", c, gf_m-1) ;
               fprintf(stream, "begin\n") ; }
           else {
               fprintf(stream, "--\n") ;
               fprintf(stream, "-- c * a^%d\n", c) ;
               fprintf(stream, "--\n") ;
               fprintf(stream, "function gfmul_%d(c : std_logic_vector(%d downto 0)) return std_logic_vector is\n", c, gf_m-1) ;
               fprintf(stream, "variable gfmul_%d : std_logic_vector(%d downto 0) ;\n", c, gf_m-1) ;
               fprintf(stream, "begin\n") ;
               fprintf(stream, "    return (gfmul_%d) ;\n", c) ;
               fprintf(stream, "end gfmul_%d ;\n\n", c) ;
               return ;
           }
           break ;
       default :
           break ;
   }

   for (i=0; i<gf_m; i++) {
       mul_matrix[i] = poly_table[a] ;
       a = (a+1)%(gf_field-1) ; /* Modulo GF(2^n) is required correctly index table */
   }


   bit_mask = 0x01 ;
   bit_cnt = 0 ;

   for (i=0; i<gf_m ; i++) {
       if (hdl_type == C) fprintf(stream, "gf_mul[%d] = ",  i) ;
       if (hdl_type == VLOG) fprintf(stream, "gfmul_%d[%d] = ", c, i) ;
       if (hdl_type == VHDL) fprintf(stream, "gfmul_%d(%d) := ", c, i) ;

       for (j=0; j<gf_m; j++) { // Find number of '1' bits
           if (mul_matrix[j] & bit_mask) {
               bit_cnt += 1 ;
           }
       }
 
       for (j=0; j<gf_m; j++) { // print resulting line for each '1' bit
           if (mul_matrix[j] & bit_mask && bit_cnt > 1) {
              if ((hdl_type == VLOG) || (hdl_type == C)) fprintf(stream, "c[%d]", j) ;
              if (hdl_type == VHDL) fprintf(stream, "c(%d)", j) ;
                  if ((hdl_type == VLOG) || (hdl_type == C)) 
                      fprintf(stream, " ^ ") ;
                  else if (hdl_type == VHDL) 
                      fprintf(stream, " xor ") ;
                  bit_cnt -= 1 ;
               } else if ((mul_matrix[j] & bit_mask) && bit_cnt == 1) {
                   if ((hdl_type == VLOG) || (hdl_type == C)) fprintf(stream, "c[%d]", j) ;
                   if (hdl_type == VHDL) fprintf(stream, "c(%d)", j) ;
                   fprintf(stream, " ; ") ;
                   bit_cnt -= 1 ;
               }
       }
        
       fprintf(stream, "\n") ;
       bit_mask = bit_mask << 1 ;
   }
   switch (hdl_type) {
       case C:
           fprintf(stream, "r = 0 ;\n") ;
           fprintf(stream, "for ( i=0; i<%d; i++) {\n", gf_m) ;  
           fprintf(stream, "  r = ((gf_mul[i] & 0x01) << i) | r ;\n") ;
           fprintf(stream, "}\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "return (r) ;\n") ;
           fprintf(stream, "}\n\n") ;
           break;
       case VLOG :
           fprintf(stream, "end\n") ;
           fprintf(stream, "endfunction\n") ;
           break;
       case VHDL :
           fprintf(stream, "    return (gfmul_%d) ;\n", c) ;
           fprintf(stream, "end gfmul_%d ;\n", c) ;
           break ;
       default :
           break ;
   }

}
//
// Generate Reed-Solomon hardware multiplier - multiplies  a*b in GF(2^n)
//
// gen_rs_mul()
//
// input:
//     gf_m: RS 2^m, where m is the feild
//     hdl_type : HDL type, 0 = vhdl, 1 = verilog, 2 = C
//
// output:
//     HDL code
//
// returns:
//     void
//
//     
void gen_rs_mul(FILE *stream, int gf_m, unsigned int hdl_type)
{

int i, j, k, t;

int bit_cnt ;
unsigned char bit_mask ;

    i = 0 ;
    k = 0 ;
    t = 0 ;

   switch (hdl_type) {
       case C:
            fprintf(stream, "/* Multiply a feild value a * b */\n") ;
            fprintf(stream, "/* T[i] terms are for GF(2^8) */\n") ; 
            fprintf(stream, "/* gf_mul[i] terms are dependent on the polynomial */\n") ;
            fprintf(stream, "unsigned char gfmul_ab( c, d )\n") ;
            fprintf(stream, "unsigned char c, d ;\n") ;
            fprintf(stream, "{\n") ;
            fprintf(stream, "int i ;\n") ;
            fprintf(stream, "unsigned char a[8], b[8], t[14], gf_mul[8], r ;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "  for ( i=0; i<%d; i++) {  /* Convert to single bit array for multiply */\n", gf_m) ;
            fprintf(stream, "    a[i] = c & 0x01 ;\n") ;
            fprintf(stream, "    c = c >> 1 ;\n") ;
            fprintf(stream, "  }\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "  for ( i=0; i<%d; i++) {  /* Convert to single bit array for multiply */\n", gf_m) ;
            fprintf(stream, "    b[i] = d & 0x01 ;\n") ;
            fprintf(stream, "    d = d >> 1 ;\n") ;
            fprintf(stream, "  }\n") ;
            fprintf(stream, "\n") ;  
            fprintf(stream, "  for ( i=0; i<%d; i++) \n", gf_m) ;
            fprintf(stream, "    gf_mul[i] = 0 ;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "  for (i=0; i <= %d ; i++ )\n", 2*gf_m-2) ;
            fprintf(stream, "    t[i] = 0 ;\n") ;
            fprintf(stream, "\n") ;
           break ;

       case VLOG : 

           fprintf(stream, "function[%d:0] gfmul_ab  ;\n", gf_m-1) ;
           fprintf(stream, "input[%d:0] a ;\n", gf_m-1) ;
           fprintf(stream, "input[%d:0] b ;\n", gf_m-1) ;
           fprintf(stream, "reg[%d:0] t ;\n", 2*gf_m-1) ;
           fprintf(stream, "begin\n") ;

           break;

       case VHDL :
            fprintf(stream, "--\n") ;
            fprintf(stream, "-- GF multiply a*b\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "function gfmul_ab (a : std_logic_vector(%d downto 0); b : std_logic_vector(%d downto 0)) return std_logic_vector is\n", gf_m-1, gf_m-1) ;
            fprintf(stream, "variable t: std_logic_vector(%d downto 0) ;\n",2*gf_m-2);
            fprintf(stream, "variable gfmul_ab: std_logic_vector(%d downto 0) ;\n", gf_m-1) ;
            fprintf(stream, "begin\n") ;
           break ;
       default :
           break ;
   }

    while (t < gf_m) {
        k = t ;
        j = t ;
       switch (hdl_type) {
           case C:
           case VLOG : 
               fprintf(stream, "t[%d] = ", t) ;
               break;

           case VHDL :
               fprintf(stream, "t(%d) := ", t) ;
               break ;
           default :
               break ;
       }

        while (k >= 0) {
            if (i == t ) {
                if (hdl_type == VLOG || hdl_type == C) fprintf(stream, "(a[%d]&b[%d])", i, j) ;
                if (hdl_type == VHDL) fprintf(stream, "(a(%d) and b(%d))", i, j) ;
            }
            else {
                if (hdl_type == VLOG || hdl_type == C) fprintf(stream, "(a[%d]&b[%d]) ^ ", i, j) ;
                if (hdl_type == VHDL) fprintf(stream, "(a(%d) and b(%d)) xor ", i, j) ;
            }
            i = i + 1 ;
            j = j - 1 ;
            k = k - 1 ;
        }
        fprintf(stream, " ;\n") ;
        i = 0 ;
          t = t + 1 ;
    } 

    t = gf_m ;
    i = 1 ;
    k = 1 ;
    while (t < 2*gf_m-1) {
        j = gf_m - 1 ;
        i = k ;
       switch (hdl_type) {
           case C:
           case VLOG : 
               fprintf(stream, "t[%d] = ", t) ;
               break;

           case VHDL :
               fprintf(stream, "t(%d) := ", t) ;
               break ;
           default :
               break ;
       }
        while (i < gf_m) {
            if (i == gf_m-1){
                if (hdl_type == VLOG || hdl_type == C) fprintf(stream, "(a[%d]&b[%d])", i, j) ;
                if (hdl_type == VHDL) fprintf(stream, "(a(%d) and b(%d))", i, j) ;
            }
            else {
                if (hdl_type == VLOG || hdl_type == C) fprintf(stream, "(a[%d]&b[%d]) ^ ", i, j) ;
                if (hdl_type == VHDL) fprintf(stream, "(a(%d) and b(%d)) xor ", i, j) ;
            }
            j = j - 1 ;
            i = i + 1 ;
        }
        fprintf(stream, " ;\n") ;
        t = t + 1 ;
        k = k + 1 ;
    }

    // load multiplication matrix with x^i mod P(x) from table
    for (i=0; i<gf_m-1; i++) {
        mul_matrix[i] = poly_table[gf_m + i] ;
        //fprintf(stream, "%.x \n", mul_matrix[i]) ;
    }

    // perform matrix multiplication and print formated code string
   bit_mask = 0x01 ;
   for (i=0; i<gf_m ; i++) {

       bit_cnt = 0 ;

       switch (hdl_type) {
           case C:
               fprintf(stream, "gf_mul[%d] = ", i) ;
               break ;
           case VLOG:
               fprintf(stream, "gfmul_ab[%d] = ", i) ;
               break ;
           case VHDL:
               fprintf(stream, "gfmul_ab(%d) := ", i) ;
               break ;
       }

       switch (hdl_type) {
           case C:
           case VLOG:
               fprintf(stream, "t[%d] ^ ", i) ;  
               break ;
           case VHDL:
               fprintf(stream, "t(%d) xor ", i) ; 
               break ;
       }


       for (j=0; j<gf_m-1; j++) { // Find number of '1' bits
           if (mul_matrix[j] & bit_mask) {
               bit_cnt += 1 ;
           }
       }

       for (j=0; j<gf_m-1; j++) { // print resulting line for each '1' bit

           if (mul_matrix[j] & bit_mask && bit_cnt > 1) {
               if ((hdl_type == VLOG) || (hdl_type == C))
                   fprintf(stream, "t[%d]", j+gf_m) ;
               else if  (hdl_type == VHDL)
                   fprintf(stream, "t(%d)", j+gf_m) ;
               if ((hdl_type == VLOG) || (hdl_type == C)) 
                   fprintf(stream, " ^ ") ;
               else if (hdl_type == VHDL) 
                   fprintf(stream, " xor ") ;
               bit_cnt -= 1 ;
           } else if ((mul_matrix[j] & bit_mask) && bit_cnt == 1) {
               if ((hdl_type == VLOG) || (hdl_type == C))
                   fprintf(stream, "t[%d]", j+gf_m) ;
               else if  (hdl_type == VHDL)
                   fprintf(stream, "t(%d)", j+gf_m) ;
               fprintf(stream, " ; ") ;
               bit_cnt -= 1 ;
           }
       }
       fprintf(stream, "\n") ;
       bit_mask = bit_mask << 1 ;
   }
   switch (hdl_type) {
       case C:
           fprintf(stream, "r = 0 ;\n") ;
           fprintf(stream, "for ( i=0; i<%d; i++) {\n", gf_m) ;  
           fprintf(stream, "  r = ((gf_mul[i] & 0x01) << i) | r ;\n") ;
           fprintf(stream, "}\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "return (r) ;\n") ;
           fprintf(stream, "}\n\n") ;
           break ;
       case VLOG : 
           fprintf(stream, "end\n") ;
           fprintf(stream, "endfunction\n") ;
           break;

       case VHDL :
           fprintf(stream, "    return (gfmul_ab) ;\n") ;
           fprintf(stream, "end gfmul_ab ;\n") ;
           break ;
       default :
           break ;
   }

}

void gen_gfarith_hdr(FILE *stream, int hdl_type)
{
   switch (hdl_type) {
       case C:
       case VLOG : 
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      gf_arith.c                                    //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           break;

       case VHDL :
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--                  Generated by RSTK                                 --\n") ;  
           fprintf(stream, "--                      gf_arith.vhd                                  --\n") ;  
           fprintf(stream, "--                                                                    --\n") ;  
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "LIBRARY IEEE ;\n") ;
           fprintf(stream, "use IEEE.std_logic_1164.all ;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use     IEEE.std_logic_unsigned.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use     IEEE.NUMERIC_STD.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "package gf_arith is\n\n") ;
           break ;
       default :
           break ;
   }
}

void gen_scalar_proto(FILE *stream, unsigned int coeff, unsigned int gf_m, unsigned int gf_field, unsigned int hdl_type)
{
int a, c ;

   a = (unsigned int)log_poly_table[coeff]%(gf_field-1) ;
   c = a ;

   switch (hdl_type) {
       case C:
           fprintf(stream, "unsigned char gfmul_%d( unsigned char ) ;\n", c) ;
           break ;
       case VLOG : 
           break;
       case VHDL :
           fprintf(stream, "function gfmul_%d(c : std_logic_vector(%d downto 0)) return std_logic_vector ;\n", c, gf_m-1) ;
           break ;
       default :
           break ;
   }

}
void gen_multiply_proto(FILE *stream, unsigned int gf_m, unsigned int hdl_type)
{

   switch (hdl_type) {
       case C:
           fprintf(stream, "unsigned char gfmul_ab( unsigned char, unsigned char ) ;\n") ;
           break ;
       case VLOG : 
           break;
       case VHDL :
           fprintf(stream, "function gfmul_ab(a: std_logic_vector(%d downto 0); \n", gf_m-1) ;
           fprintf(stream, "                  b: std_logic_vector(%d downto 0)) return std_logic_vector ;\n", gf_m-1) ;
           break ;
       default :
           break ;
   }

}

void gen_proto_footer(FILE *stream, unsigned int hdl_type)
{

   switch (hdl_type) {
       case C:
       case VLOG : 
           break;
       case VHDL :
           fprintf(stream, "end package gf_arith ;\n\n") ;
           fprintf(stream, "package body gf_arith is\n") ;
           break ;
       default :
           break ;
   }

}

void gen_gfarith_footer(FILE *stream, unsigned int hdl_type)
{

   switch (hdl_type) {
       case C:
       case VLOG : 
           fprintf(stream, "\n\n//END gf_arith ;\n") ;
           break;
       case VHDL :
           fprintf(stream, "\n\nEND gf_arith ;\n") ;
           break ;
       default :
           break ;
   }

}

