/*******************************************************************************/
/*                                                                             */
/*                         RS_TK                                               */
/*                   Reed-Solomon Tool Kit                                     */
/*   Generates 'C', VHDL, or Verilog files for RS codes from GF(2^3)           */
/*   to GF(2^*). These files can be compiled using either 'C',VHDL,            */
/*   or Verilog compilers from a variety of vendors to produce                 */
/*   synthesizable or simulatable logic. The intent is to produce              */
/*   a machine/platform independant tool for RS codes.                         */ 
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*   usage:                                                                    */
/*           rstk -p -n -c -t -g -f -r                                         */
/*           -p<polynomial dvb,itu,x4,x5,"101101110">                          */
/*           -n<2^n>            n=1,2,3,4,5,6,7,8                              */
/*           -c<1,2,3,...    >  number of check symbols                        */
/*           -b<integer>        polynomial base                                */
/*           -g<vhdl, vlog, c>  code generation                                */
/*           -f<alt,xlx>        FPGA mmanufacturer                             */
/*           -r<syn,sim>        ROM type  Synthesis or Simulation              */
/*                                                                             */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */ 
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/
#include <stdio.h>
#include <string.h>

#include "rsccg.h"
#include "rsccg_proto.h"
#include "argtable.h"

#define VHDL     0 
#define VLOG     1 
#define C        2 

#define X_UCF    0 
#define X_COE    1 
#define ALT_MIF  2 

extern unsigned char poly_table[] ;        // Alpha GF Feild (a^n)
extern unsigned char inv_poly_table[] ;    // Inverse Alpha (1/a^n)
extern unsigned char pow_poly_table[] ;    // (a^n)^n (feild from 0 to n-1) used if base > 0
extern unsigned char log_poly_table[] ;    // Log(alpha), this table look up returns alpha from n

unsigned int p1[2] = {0x01,1} ;   // Initial polynomial is a^0 + 1*x^1
unsigned int p_wsp[256] ;         // Polynomial work space
//unsigned char gx_dup[64] ;        // G(n) duplicate array 
unsigned char gx_dup_list[1024] ; // G(n) duplicate list, entries are used to check
                                  // if a duplicate of g(n) is found 
 
FILE *rs_ccg ;          // C Array of RS Tables
//
// Files
//
char  ROM_C_TABLES[]   = {"rom_c_tables.c"} ;
char  ROM_ALPHA[]     = {"gf_alpha_rom.hex"} ;
char  ROM_INV_ALPHA[] = {"gf_inv_rom.hex"} ;
char  ROM_POW_ALPHA[] = {"gf_pow_rom.hex"} ;
char  C_GF_ALPHA_TBL[] = {"gf_alpha_tbl"} ;
char  C_GF_INVALPHA_TBL[] = {"gf_invalpha_tbl"} ;
char  C_GF_POWALPHA_TBL[] = {"gf_powalpha_tbl"} ;

char  GF_ARITH_VHDL[] = {"gf_arith.vhd"} ;
char  GF_ARITH_VLOG[] = {"gf_arith.v"} ;
char  GF_ARITH_C[]    = {"gf_arith.c"} ;

char  X_UCF_ROM_ALPHA[]    = {"x_rom_alpha.ucf"} ;
char  X_UCF_ROM_INVALPHA[] = {"x_rom_invalpha.ucf"} ;
char  X_UCF_ROM_POWALPHA[] = {"x_rom_powalpha.ucf"} ;

char  X_COE_ROM_ALPHA[]    = {"x_rom_alpha.coe"} ;
char  X_COE_ROM_INVALPHA[] = {"x_rom_invalpha.coe"} ;
char  X_COE_ROM_POWALPHA[] = {"x_rom_powalpha.coe"} ;

char  ALT_MIF_ROM_ALPHA[]     = {"alt_rom_alpha.mif"} ;
char  ALT_MIF_ROM_INVALPHA[]  = {"alt_rom_invalpha.mif"} ;
char  ALT_MIF_ROM_POWALPHA[]  = {"alt_rom_powalpha.mif"} ;

char  RS_ENCODER_C[]    = {"rs_enc.c"} ;
char  RS_ENCODER_VLOG[] = {"rs_enc.v"} ;
char  RS_ENCODER_VHDL[] = {"rs_enc.vhd"} ;

char  RS_SYNDEC_C[]    = {"rs_syndec.c"} ;
char  RS_SYNDEC_VLOG[] = {"rs_syndec.v"} ;
char  RS_SYNDEC_VHDL[] = {"rs_syndec.vhd"} ;

char  RS_BERL_C[]    = {"rs_berl.c"} ;
char  RS_BERL_VLOG[] = {"rs_berl.v"} ;
char  RS_BERL_VHDL[] = {"rs_berl.vhd"} ;

char  RS_CHIEN_C[]    = {"rs_chien.c"} ;
char  RS_CHIEN_VLOG[] = {"rs_chien.v"} ;
char  RS_CHIEN_VHDL[] = {"rs_chien.vhd"} ;

char  RS_FORNEY_C[]    = {"rs_forney.c"} ;
char  RS_FORNEY_VLOG[] = {"rs_forney.v"} ;
char  RS_FORNEY_VHDL[] = {"rs_forney.vhd"} ;

char  RS_POLYREG_C[]    = {"rs_polymath.c"} ;
char  RS_POLYREG_VLOG[] = {"rs_polyreg.v"} ;
char  RS_POLYREG_VHDL[] = {"rs_polyreg.vhd"} ;

char  RS_ROM_VHDL[] = {"rs_rom.vhd"} ;
char  RS_ROM_VLOG[] = {"rs_rom.v"} ;
//
//  Arg table ARGC, ARGC variables
//
char  rs_poly[32] ;      // Reed Solomon Polynomial
char  rom_type[4] ;      // ROM Type 
char  fpga_mfg[4] ;      // FPGA Manufacurer
char  code_gen[4] ;      // code generator
int   chk_sym ;          // Number of codewords to check/correct
int   gf_2n ;            // GF feild 2^n 
int   base;              // Gererator polynomial base (x-a^base)

int main(int argc, char **argv)
{

    int r_deg;
    int i,n ;
    int str_test;
    int gf_poly, gf_field ;
    int code_type ;

/*           rstk -p -n -c -t -g -f -r                                */
/*           -p<polynomial dvb,itu,x4,x5,"101101110">                 */
/*           -n<255,32,16,8>                                          */
/*           -c<2,3,...    >    number of check symbols               */
/*           -b<integer>        polynomial base                       */
/*           -g<vhdl, vlog, c>  code generation                       */
/*           -f<alt,xlx>        FPGA mmanufacturer                    */
/*           -r<syn,sim>        ROM type  Synthesis or Simulation     */
/*                                                                    */
    arg_rec argtable[] =
      {
      {"-p", "<rs_poly>",       arg_str, rs_poly,   "x4",   "\t RS Polynomial"},
      {"-n", "<gf_feild_size>", arg_int, &gf_2n,    "4",    "\t GF Field"},
      {"-c", "<check_symbols>", arg_int, &chk_sym,  "1",    "\t RS Check Symbols"},
      {"-b", "<poly_base>",     arg_int, &base,     "0",    "\t Polynomial Base"},
      {"-g", "<code_gen>",      arg_str, code_gen,  "vhdl", "\t Code Generator"},
      {"-f", "<fpga_mfg>",      arg_str, fpga_mfg,  "xlx",  "\t FPGA Manufacturer"},
      {"-r", "<rom_type>",      arg_str, rom_type,  "sim",  "\t ROM Type"}
      };
   const size_t narg = sizeof(argtable)/sizeof(arg_rec);

   /*-- process the command line args --*/
   if (argc==1)
      {
      /*-- display program usage and exit. --*/
      printf("Usage: %s %s\n", argv[0], arg_syntax(argtable,narg));
      printf("%s\n",arg_glossary(argtable,narg,"  "));
      return 0;
      }
   else
      {
      /*-- scan command line arguments from argv[] --*/
      char cmdline[200], errmsg[200], errmark[200];
      if (!arg_scanargv(argc, argv, argtable, narg, cmdline, errmsg, errmark))
         {
         /*-- arg error occurred, print error message and exit --*/
         printf("ERROR: %s\n", cmdline);
         printf("       %s %s\n", errmark, errmsg);
         return (1);
         }
      }

   printf("\n") ;
   printf("Reed Solomon Polynomial: %s\n", rs_poly) ;
   printf("GF Field Size: %d\n", gf_2n) ;
   printf("Number of check symbols: %d\n", chk_sym) ;
   printf("Generator Polynomial Base: %d\n", base) ;
   printf("Code Generator: %s\n", code_gen) ;
   printf("FPGA Manufacturer: %s\n", fpga_mfg) ;
   printf("ROM Type: %s\n", rom_type) ;

   // Build/Generate GF Tables
   str_test = strcmp(rs_poly, "x4");
   if (str_test == 0) {
       if (gf_2n != 4) {
           printf("Error - Field Size Is Not Equal to Feild Generator Polynomial Degree\n") ;
           return(0) ;
       }
       gf_poly = GF2_E4 ;
       rsgen_gf_tables(gf2_e4,gf_2n,base,0) ;
   } else {
          str_test = strcmp(rs_poly, "x5");
          if (str_test == 0) {
              if (gf_2n != 5) {
                  printf("Error - Field Size Is Not Equal to Feild Generator Polynomial Degree\n") ;
                  return(0) ;
              }
              gf_poly = GF2_E5 ;
              rsgen_gf_tables(gf2_e5,gf_2n,base,0) ;
          } else {
                  str_test = strcmp(rs_poly, "dvb");
                  if (str_test == 0) {
                      if (gf_2n != 8) {
                          printf("Error - Field Size Is Not Equal to Feild Generator Polynomial Degree\n") ;
                          return(0) ;
                      }
                      gf_poly = GF2_E8_dvb ;
                      rsgen_gf_tables(gf2_e8_dvb,gf_2n,base,0) ;
                  } else {
                          str_test = strcmp(rs_poly, "itu");
                          if (str_test == 0) {
                              if (gf_2n != 8) {
                                  printf("Error - Field Size Is Not Equal to Feild Generator Polynomial Degree\n") ;
                                  return(0) ;
                              }
                              gf_poly = GF2_E8_itu ;
                              rsgen_gf_tables(gf2_e8_itu,gf_2n,base,0) ;
                          } else {
                              gf_poly = bstr_to_int(rs_poly) ;
                              rev_str(rs_poly) ;
                              rsgen_gf_tables(rs_poly,gf_2n,base,0) ;
                          }
                      }
                        
          }
       }
//
// Determine Code Generation model
//
   str_test = strcmp(code_gen, "vhdl");
   if (str_test == 0) code_type = VHDL ;
   else {
        str_test = strcmp(code_gen, "vlog");
        if (str_test == 0)  code_type = VLOG ;
        else {
              str_test = strcmp(code_gen, "C");
              if (str_test == 0) code_type = C ;
              else {
                   printf("ERROR - Invalid Code Type\n") ;    
                   return (1) ;
              }
        }
   }

   // Compute 2^gf_m
   gf_field = 1 ;
   for (n=0; n < gf_2n; n++) gf_field = 2*gf_field ;
//
//
//  Expand code generator polynomial (x-a^base)(x-a^base+1)...
//
//
    r_deg = expand_gen_poly(gf_poly, chk_sym, base, gf_2n, gf_field) ;
//
//
//
   // Open File for GF_ARITH and Generate GF_ARITH.XXX
    switch (code_type) {
        case C:
            rs_ccg = fopen(GF_ARITH_C, "w") ;
            gen_gfarith_hdr(rs_ccg, C) ;
            break ;
        case VLOG:
            rs_ccg = fopen(GF_ARITH_VLOG, "w") ;
            gen_gfarith_hdr(rs_ccg, VLOG) ;
            break ;
        case VHDL:
            rs_ccg = fopen(GF_ARITH_VHDL, "w") ;
            gen_gfarith_hdr(rs_ccg, VHDL) ;
            break ;
    }
    //
    // Find duplicates of generator coeffiecients and output only different g(n)
    // Output PROTOTYPES for C, VHDL. VLOG does not create a prototype

    // Clear entries 
    for (i=0; i < 1024; i++)  
        gx_dup_list[i] = 0 ;

    // Write generator coeffiecient proto types to file
    for (i=0; i <= r_deg; i++ ) {
		if (gx_dup_list[p_wsp[i]] == 0) {
          gen_scalar_proto(rs_ccg, p_wsp[i], gf_2n, gf_field, code_type) ;
          gx_dup_list[p_wsp[i]] = p_wsp[i] ;
		}
    }
    // Write generator coeff's for zero's of generator polynomial
    // Add to duplicate list the value if there is no entry
      for (i=base; i < base+r_deg; i++) { 
          if (gx_dup_list[poly_table[i]] == 0) {
              gx_dup_list[poly_table[i]] = poly_table[i] ;
              gen_scalar_proto(rs_ccg,poly_table[i], gf_2n, gf_field, code_type) ;
          }
      }

    // Write general field muliplyer
    gen_multiply_proto(rs_ccg, gf_2n, code_type) ;

    // Output constant multiplier for chien search 
      for (i=1; i < r_deg; i++)
          if (gx_dup_list[poly_table[i]] == 0)
              gen_scalar_proto(rs_ccg,poly_table[i], gf_2n, gf_field, code_type) ;

    gen_proto_footer(rs_ccg, code_type) ;

    // Clear entries 
    for (i=0; i < 1024; i++)  
        gx_dup_list[i] = 0 ;


    //
    // Generate code for constant multiplier for each term of G(x)
    //
    for (i=0; i < r_deg; i++ ) {
		if (gx_dup_list[p_wsp[i]] == 0) {
			gen_rs_scalar_mul(rs_ccg, p_wsp[i], gf_2n, gf_field, code_type) ;
            gx_dup_list[p_wsp[i]] = p_wsp[i] ;
		}
    }
    // Write general multipler for G(x)
    gen_rs_mul(rs_ccg, gf_2n, code_type) ;

    // Write constant multipler for each Zero of G(X)
    for (i=base; i < base+r_deg; i++) {
          if (gx_dup_list[poly_table[i]] == 0) {
              gx_dup_list[poly_table[i]] = poly_table[i] ;
              gen_rs_scalar_mul(rs_ccg,poly_table[i], gf_2n, gf_field, code_type) ;
         } 
    }

    // Output constant multiplier for chien search
    for (i=0; i < r_deg; i++) 
        if (gx_dup_list[poly_table[i]] == 0)
            gen_rs_scalar_mul(rs_ccg,poly_table[i], gf_2n, gf_field, code_type) ;

    gen_gfarith_footer(rs_ccg, code_type) ;

    fclose(rs_ccg) ;

    //
    // Create ROM files and initialization images
    //
    switch (code_type) {
        case C:
            rs_ccg = fopen(ROM_C_TABLES, "w") ;
            gen_rs_rom_data(rs_ccg, poly_table, C_GF_ALPHA_TBL,  C, gf_field) ;
            gen_rs_rom_data(rs_ccg, inv_poly_table, C_GF_INVALPHA_TBL,  C,
            gf_field) ;
            if (base >0 ) gen_rs_rom_data(rs_ccg, pow_poly_table, C_GF_POWALPHA_TBL,  C, gf_field) ;
            fclose(rs_ccg) ;
            break ;

        case VLOG:
            rs_ccg = fopen(ROM_ALPHA, "w") ;
            gen_rs_rom_data(rs_ccg, poly_table, ROM_ALPHA,  VLOG, gf_field) ;
            fclose(rs_ccg) ;
            rs_ccg = fopen(ROM_INV_ALPHA, "w") ;
            gen_rs_rom_data(rs_ccg, inv_poly_table, ROM_INV_ALPHA,  VLOG, gf_field) ;
            fclose(rs_ccg) ;
            if (base >0 ) {
                rs_ccg = fopen(ROM_POW_ALPHA, "w") ;
				gen_rs_rom_data(rs_ccg, pow_poly_table, ROM_POW_ALPHA,  VLOG, gf_field) ;
                fclose(rs_ccg) ;
			}
            break ;

        case VHDL:
            rs_ccg = fopen(ROM_ALPHA, "w") ;
            gen_rs_rom_data(rs_ccg, poly_table, ROM_ALPHA,  VHDL, gf_field) ;
            fclose(rs_ccg) ;
            rs_ccg = fopen(ROM_INV_ALPHA, "w") ;
            gen_rs_rom_data(rs_ccg, inv_poly_table, ROM_INV_ALPHA,  VHDL, gf_field) ;
            fclose(rs_ccg) ;
            if (base >0 ) {
                rs_ccg = fopen(ROM_POW_ALPHA, "w") ;
				gen_rs_rom_data(rs_ccg, pow_poly_table, ROM_POW_ALPHA,  VHDL, gf_field) ;
                fclose(rs_ccg) ;
			}
            break ;
    }

   // Xilinx
         str_test = strcmp(fpga_mfg, "xlx");
         if (str_test == 0) {
             rs_ccg = fopen(X_UCF_ROM_ALPHA, "w") ;
             gen_rom_init(rs_ccg, poly_table, X_UCF_ROM_ALPHA,  X_UCF, gf_field) ;
             fclose(rs_ccg) ;
             rs_ccg = fopen(X_UCF_ROM_INVALPHA, "w") ;
             gen_rom_init(rs_ccg, inv_poly_table, X_UCF_ROM_INVALPHA,  X_UCF, gf_field) ;
             fclose(rs_ccg) ;
             if (base > 0) {
                 rs_ccg = fopen(X_UCF_ROM_POWALPHA, "w") ;
                 gen_rom_init(rs_ccg, pow_poly_table, X_UCF_ROM_POWALPHA,  X_UCF, gf_field) ;
                 fclose(rs_ccg) ;
             }
             rs_ccg = fopen(X_COE_ROM_ALPHA, "w") ;
             gen_rom_init(rs_ccg, poly_table, X_COE_ROM_ALPHA,  X_COE, gf_field) ;
             fclose(rs_ccg) ;
             rs_ccg = fopen(X_COE_ROM_INVALPHA, "w") ;
             gen_rom_init(rs_ccg, inv_poly_table, X_COE_ROM_INVALPHA,  X_COE, gf_field) ;
             fclose(rs_ccg) ;
             if (base > 0) {
                 rs_ccg = fopen(X_COE_ROM_POWALPHA, "w") ;
                 gen_rom_init(rs_ccg, pow_poly_table, X_COE_ROM_POWALPHA,  X_COE, gf_field) ;
                 fclose(rs_ccg) ;
             }
		 }
         // Altera
         str_test = strcmp(fpga_mfg, "alt");
         if (str_test == 0) {
             rs_ccg = fopen(ALT_MIF_ROM_ALPHA, "w") ;
             gen_rom_init(rs_ccg, poly_table, ALT_MIF_ROM_ALPHA,  ALT_MIF, gf_field) ;
             fclose(rs_ccg) ;
             rs_ccg = fopen(ALT_MIF_ROM_INVALPHA, "w") ;
             gen_rom_init(rs_ccg, inv_poly_table, ALT_MIF_ROM_INVALPHA,  ALT_MIF, gf_field) ;
             fclose(rs_ccg) ;
             if (base > 0) {
                 rs_ccg = fopen(ALT_MIF_ROM_POWALPHA, "w") ;
                 gen_rom_init(rs_ccg, pow_poly_table, ALT_MIF_ROM_POWALPHA,  ALT_MIF, gf_field) ;
                 fclose(rs_ccg) ;
             }
         }

         str_test = strcmp(rom_type, "sim");
         if (str_test == 0) {
			 switch (code_type) {
			     case C:
                     break ;
			     case VLOG:
                     rs_ccg = fopen(RS_ROM_VLOG, "w") ;
                     gen_rs_rom_rtl(rs_ccg, VLOG) ;
                     fclose(rs_ccg) ;
                     break ;
			     case VHDL:
                     rs_ccg = fopen(RS_ROM_VHDL, "w") ;
                     gen_rs_rom_rtl(rs_ccg, VHDL) ;
                     fclose(rs_ccg) ;
                     break ;
			 }
         }

    // 
    // Now start generating RS Codec code
    //
    //
    // RS Encoder
    //
    switch (code_type) {
        case C:
            rs_ccg = fopen(RS_ENCODER_C, "w") ;
            gen_rs_encoder(rs_ccg, p_wsp, gf_2n, r_deg, C);
            fclose(rs_ccg) ;
            break ;
        case VLOG:
            rs_ccg = fopen(RS_ENCODER_VLOG, "w") ;
            gen_rs_encoder(rs_ccg, p_wsp, gf_2n, r_deg, VLOG) ;
            fclose(rs_ccg) ;
            break ;
        case VHDL:
            rs_ccg = fopen(RS_ENCODER_VHDL, "w") ;
            gen_rs_encoder(rs_ccg, p_wsp, gf_2n, r_deg,VHDL) ;
            fclose(rs_ccg) ;  
            break ;
    }
    //
    // RS Receiver Syndrome Decode
    //
    // gen_rs_syndec()
    //
    switch (code_type) {
        case C:
            rs_ccg = fopen(RS_SYNDEC_C, "w") ;
            gen_rs_syndec(rs_ccg, base, gf_2n, r_deg, C);
            fclose(rs_ccg) ;
            break ;
        case VLOG:
            rs_ccg = fopen(RS_SYNDEC_VLOG, "w") ;
            gen_rs_syndec(rs_ccg, base, gf_2n, r_deg, VLOG) ;
            fclose(rs_ccg) ;
            break ;
        case VHDL:
            rs_ccg = fopen(RS_SYNDEC_VHDL, "w") ;
            gen_rs_syndec(rs_ccg, base, gf_2n, r_deg, VHDL) ;
            fclose(rs_ccg) ;  
            break ;
    }
    //
    // RS Berlenkamp-Massey Error Locator
    // gen_rs_berl()
    //
    switch (code_type) {
        case C:
            rs_ccg = fopen(RS_BERL_C, "w") ;
            gen_rs_berl(rs_ccg, base, gf_2n, r_deg, chk_sym, C, gf_field);
            fclose(rs_ccg) ;
            rs_ccg = fopen(RS_POLYREG_C, "w") ;
            gen_rs_polyreg(rs_ccg, base, gf_2n, r_deg, chk_sym, C);
            fclose(rs_ccg) ;
            break ;
        case VLOG:
            rs_ccg = fopen(RS_BERL_VLOG, "w") ;
            gen_rs_berl(rs_ccg, base, gf_2n, r_deg, chk_sym, VLOG, gf_field) ;
            fclose(rs_ccg) ;
            rs_ccg = fopen(RS_POLYREG_VLOG, "w") ;
            gen_rs_polyreg(rs_ccg, base, gf_2n, r_deg, chk_sym, VLOG);
            fclose(rs_ccg) ;
            break ;
        case VHDL:
            rs_ccg = fopen(RS_BERL_VHDL, "w") ;
            gen_rs_berl(rs_ccg, base, gf_2n, r_deg, chk_sym, VHDL, gf_field) ;
            fclose(rs_ccg) ;
            rs_ccg = fopen(RS_POLYREG_VHDL, "w") ;
            gen_rs_polyreg(rs_ccg, base, gf_2n, r_deg, chk_sym, VHDL);
            fclose(rs_ccg) ;            
            break ;
    }
    //
    // RS Chien Root Search
    // gen_rs_chien()
    //
    switch (code_type) {
        case C:
            rs_ccg = fopen(RS_CHIEN_C, "w") ;
            gen_rs_chien(rs_ccg, base, gf_2n, r_deg, chk_sym, C);
            fclose(rs_ccg) ;
            break ;
        case VLOG:
            rs_ccg = fopen(RS_CHIEN_VLOG, "w") ;
            gen_rs_chien(rs_ccg, base, gf_2n, r_deg, chk_sym, VLOG) ;
            fclose(rs_ccg) ;
            break ;
        case VHDL:
            rs_ccg = fopen(RS_CHIEN_VHDL, "w") ;
            gen_rs_chien(rs_ccg, base, gf_2n, r_deg, chk_sym, VHDL) ;
            fclose(rs_ccg) ;  
            break ;
    } 
    //
    // RS Error Correction Value
    // gen_rs_forney()
    //
    switch (code_type) {
        case C:
            rs_ccg = fopen(RS_FORNEY_C, "w") ;
            gen_rs_forney(rs_ccg, base, gf_2n, r_deg, chk_sym, C, gf_field);
            fclose(rs_ccg) ;
            break ;
        case VLOG:
            rs_ccg = fopen(RS_FORNEY_VLOG, "w") ;
            gen_rs_forney(rs_ccg, base, gf_2n, r_deg, chk_sym, VLOG, gf_field) ;
            fclose(rs_ccg) ;
            break ;
        case VHDL:
            rs_ccg = fopen(RS_FORNEY_VHDL, "w") ;
            gen_rs_forney(rs_ccg, base, gf_2n, r_deg, chk_sym, VHDL, gf_field) ;
            fclose(rs_ccg) ;  
            break ;
    } 
    //
    // RS Decoder - Error Correction
    // gen_rs_decoder()
    return(0) ;

}
